import requests

REQUESTS_METHODS = {
    "GET": requests.get,
    "POST": requests.post,
    "PUT": requests.put,
    "PATCH": requests.patch,
    "DELETE": requests.delete,
}


class RESTClient(object):
    def __init__(self, base_urn, port=80, protocol="http"):
        self._base_urn = base_urn
        self._port = port
        self._protocol = protocol

        self.base_url = "%s://%s:%s" % (self._protocol, self._base_urn, self._port)
        self._api_endpoint = None
        self._auth = None
        self._username = self._password = None

    def set_api_endpoint(self, endpoint):
        """Sets the endpoint for the API"""
        self._api_endpoint = "%s/%s" % (self.base_url, endpoint)

    def get_api_endpoint(self):
        """Returns the configured API endpoint"""
        return self._api_endpoint

    def set_authentication_basic(self, username, password):
        """Sets default authentication (username, password)"""
        self._auth = requests.auth.HTTPBasicAuth(username, password)
        self._username = username
        self._password = password

    def set_authentication_digest(self, username, password):
        self._auth = requests.auth.HTTPDigestAuth(username, password)
        self._username = username
        self._password = password

    def send_request(self, method, object_id=None, data=None, params=None, headers=None, json_content=None, files=None):
        url = self._api_endpoint
        if object_id:
            url += '/' + object_id
        if not headers:
            headers = {}

        check_ssl_certs = False
        requests.packages.urllib3.disable_warnings()
        request_response = REQUESTS_METHODS[method](url, data=data, params=params, headers=headers, json=json_content,
                                                    verify=check_ssl_certs, auth=self._auth, files=files)
        return request_response


if __name__ == "__main__":
    # TODO: Delete
    """Small tests to see if the client works. To delete in the future"""
    rest_client = RESTClient("jsonplaceholder.typicode.com", 443)
    rest_client.set_api_endpoint("posts")
    request_response = rest_client.send_request("GET")
    print(request_response)



import os

from configobj import ConfigObj

FWK_CONFIG = None


def get_slayer_configuration():
    """Gets the Slayer configuration, an object with all the settings for running Slayer.

    The Slayer configuration is stored in a global variable, and allows to modify the execution of Slayer."""
    global FWK_CONFIG
    if FWK_CONFIG is None:
        config_path = os.getenv("FWK_CONFIG")
        FWK_CONFIG = ConfigObj(config_path)
    return FWK_CONFIG





import logging
import time
import math


def dec_log_execution_time(method):
    def timed_function(*args, **kw):
        line_length = 42
        # logging.debug("{0} start".format(method.__name__).center(line_length, '-'))
        start_time = time.time()
        result = method(*args, **kw)
        end_time = time.time()
        logging.debug(
            ('{0} Total time: %2.2f sec'.format(method.__name__) % (end_time - start_time)).center(line_length, '-'))
        return result

    return timed_function


def retry_until_value(value, retries=3, delay=1):
    """Retries a function or method until it returns a specific value

        delay sets the initial delay in seconds, and backoff sets the factor by which
        the delay should lengthen after each failure. backoff must be greater or equal ot 1.
        tries must be at least 0, and delay
        greater than 0."""


def retry_until_not_none(retries=3, delay=1):
    """Retries a function or method until it returns different than None

    delay sets the initial delay in seconds, and backoff sets the factor by which
    the delay should lengthen after each failure. backoff must be greater or equal ot 1.
    tries must be at least 0, and delay
    greater than 0."""

    retries = math.floor(retries)
    if retries < 0:
        raise ValueError("tries must be 0 or greater")
    if delay <= 0:
        raise ValueError("delay must be greater than 0")

    def deco_retry(f):
        def f_retry(*args, **kwargs):
            mtries, mdelay = retries, delay  # make mutable
            retry_value = None
            while mtries > 0:
                try:
                    retry_value = f(*args, **kwargs)  # Do an attempt
                    if retry_value is not None:
                        return retry_value
                    else:
                        # TODO: Change exception
                        raise Exception("Couldn't get the element")
                except:
                    mtries -= 1  # consume an attempt
                    if mtries > 0:
                        logging.debug("Retrying...")
                    time.sleep(mdelay)  # wait...
            return retry_value  # Ran out of tries

        return f_retry  # true decorator -> decorated function

    return deco_retry  # @retry(arg[, ...]) -> true decorator


def dec_retry_bool_function(retries, wait=3, backoff=2):
    """Retries a function or method until it returns True.

    @wait sets the initial delay in seconds, and @backoff sets the factor by which
    the delay should lengthen after each failure. @backoff must be greater than 1,
    @retries must be at least 0, and delay greater than 0."""
    # TODO: Improve to use condition instead of a bool value (True)
    if backoff <= 1:
        raise ValueError("backoff must be greater than 1")

    tries = math.floor(retries)
    if tries < 0:
        raise ValueError("tries must be 0 or greater")
    if wait <= 0:
        raise ValueError("delay must be greater than 0")

    def deco_retry(f):
        def function_retry(*args, **kwargs):
            mtries, mdelay = tries, wait  # mutable values

            function_value = f(*args, **kwargs)
            while mtries > 0:
                if function_value is True:  # Done on success
                    return True
                logging.debug("Retrying...")
                mtries -= 1
                time.sleep(mdelay)
                mdelay *= backoff
                function_value = f(*args, **kwargs)  # Try again
            return False  # Ran out of tries

        return function_retry  # true decorator -> decorated function

    return deco_retry  # @retry(arg[, ...]) -> true decorator


def log_execution_time(method):
    def timed_function(*args, **kw):
        line_length = 42
        start_time = time.time()
        result = method(*args, **kw)
        end_time = time.time()
        total_execution_time = end_time - start_time
        logging.info('Execution time: %2.2f sec' % total_execution_time)
        return result

    return timed_function


# TODO: change behavior so it takes screenshots before/after function ends
# TODO: Use context in order to access the webdriver and step name
def take_screenshot(method):
    def web_function(*args, **kw):
        result = method(*args, **kw)
        # self.driver.get_screenshot_as_file(os.path.join(os.getcwd(),
        # "screenshots", screenshot_name))
        return result

    return web_function



from slayer.behave_executor import BehaveExecutor
from slayer.slayer_runner import SlayerRunner
import os
import sys

PACKAGE_PARENT = '..'
SCRIPT_DIR = os.path.dirname(os.path.realpath(os.path.join(os.getcwd(), os.path.expanduser(__file__))))
sys.path.append(os.path.normpath(os.path.join(SCRIPT_DIR, PACKAGE_PARENT)))


def run_framework(framework_name="SLAYER"):
    """Sets all the settings required for executing Slayer.

    - Configures the necessary environment variables
    -- logger
    -- output folder
    -- Slayer report
    - Sets Behave-specific variables, like the paths where the feature files will be located and tags to run
    - Calls the behave executor"""

    slayer_runner = SlayerRunner(framework_name)
    slayer_runner.parse_slayer_arguments()
    slayer_runner.set_environment_variables()

    slayer_runner.cleanup_output_folder()

    # Duplicate stdout to the slayer logger file
    slayer_runner.configure_stdout_logger()

    # configure execution for Slayer. This step needs to be executed after creating the behave config object since
    # Slayer overrides some of the settings behave sets when importing the library
    slayer_runner.configure_execution()

    # FIXME: Add variable for not printing the banner
    slayer_runner.log_slayer_information()

    # FIXME: Mention in doc that adding logging in the environment.py file could mess up the logging (for the before_all
    # FIXME: especially. Maybe a new logger could be added....
    behave_executor = BehaveExecutor()
    behave_executor.create_behave_config()
    # TODO: improve. Add if
    #behave_executor.setup_logger(slayer_runner.variables["FWK_OUTPUT_DIR"])
    behave_executor.setup_allure_logger(slayer_runner.variables["FWK_OUTPUT_DIR"])

    #behave_executor.setup_logger(slayer_runner.variables["FWK_OUTPUT_DIR"])
    #behave_executor.parse_behave_arguments()

    behave_executor.import_steps_directories()
    behave_executor.call_executor()
    # TODO: Refactor results output (execution summary)
    # TODO: Reporter Factory. If report is enabled...
    slayer_runner.generate_allure_report()


if __name__ == "__main__":
    run_framework(framework_name="SLAYER-dev")



import logging
import time
from behave import step
from selenium import webdriver

@step("I open a browser")
def step_impl(context, maximized=True):
    context.driver = webdriver.Chrome()
    if maximized:
        context.driver.maximize_window()


@step("I navigate to the Wikipedia page")
def step_impl(context):
    context.wikipedia_page = WikipediaPage(context.driver)
    logging.info("Navigating to the Wikipedia page")
    context.wikipedia_page.navigate()

@step("I log {text}")
def step_impl(context, text):
    logging.info(text)
    for i in range(20):
        logging.info("simulating logs {}...".format(i))


@step("I see in the page '{search_text}'")
def step_impl(context, search_text):
    logging.info("Searching for the text '{}'".format(search_text))
    context.wikipedia_page.search_for(search_text)
    time.sleep(1)
    
    
@step("the test fails")
def step_impl(context):
    assert False



import sys
import os

class SysStdoutLogger(object):
    """sys.stdout logger class

    This class duplicates the console output and saves it to a file, in order to consolidate
    the behave and slayer logs.
    This class is meant to overwrite the sys.stdout output"""

    def __init__(self, path, filename="output.log"):
        self.terminal = sys.stdout
        self.log = open(os.path.join(path, filename), "w")

    def __getattr__(self, attr):
        return getattr(self.terminal, attr)

    def write(self, message):
        self.terminal.write(message)
        self.log.write(message)

    def flush(self):
        pass



import os
import sys
import fnmatch
from argparse import ArgumentParser

from behave.__main__ import run_behave
from behave.configuration import Configuration as BehaveConfig
from behave.formatter.base import StreamOpener


class BehaveExecutor(object):

    def __init__(self):
        self.behave_config = None
        """Sets behave-specific arguments.

                Note: This function uses both the arguments in the behave.ini file and the command line arguments, but the latter
                take precedence."""
        # TODO: cannot send parameters
        # TODO: Mention the arguments

    def parse_behave_arguments(self):
        """Reads the Behave-specific arguments provided in the execution of Slayer."""
        parser = ArgumentParser()
        # Behave arguments
        # TODO: Complete with the rest of the arguments

    def create_behave_config(self):
        self.behave_config = BehaveConfig(sys.argv[1:])
        # TODO: Create functions to load the config files
        # cfg.environment_file = # Configurable by user

    def call_executor(self):
        """Calls the Behave executor to run the scenarios"""
        run_behave(self.behave_config)

    def setup_allure_logger(self, output_directory):
        """Inserts the Allure formatter as the first element of the outputs list, in order to make it match with the
        formats. Note that the default Behave formatter does not add a format)"""
        allure_report_directory = "allure"
        self.behave_config.outputs.insert(0, StreamOpener(filename=os.path.join(output_directory, allure_report_directory)))
        self.behave_config.format.append("allure_behave.formatter:AllureFormatter")


    @classmethod
    def import_steps_directories(cls):
        """Imports steps subdirectories.

        The steps directory, along with its subfolders are imported in order to let Behave find the steps.
        This allows Slayer to have the features and steps anywhere, regardless of where in the path it's called."""
        features_directory = os.path.join(os.path.dirname(sys.argv[0]), "features")

        # os.environ.setdefault('FEATURES_PATH', FEATURES_DIRECTORY)
        sys.path.append(features_directory)
        for dirpath, dirnames, filenames in os.walk(features_directory):
            for directory in dirnames:
                filenames = os.listdir(os.path.join(dirpath, directory))
                for python_script in fnmatch.filter(filenames, "*.py"):
                    if python_script == '__init__.py':
                        continue
                    python_module = directory + "." + python_script[:-3]
                    if dirpath != features_directory:
                        python_module = dirpath.replace(features_directory, "")[1:] + "." + python_module
                        python_module = python_module.replace(os.sep, ".")
                    __import__(python_module, locals(), globals())
        del sys.path[-1]



"""Environment file for behave
Selenium grid may be set here too.
TODO: This is deactivated for Alpha-5"""
import logging

LINE_LENGTH = 42
PASSED = "passed"
FAILED = "failed"

def before_all(context):
    # Setup logging for SLAYER, according to behave API reference:
    # http://python-behave.readthedocs.io/en/latest/api.html#logging-setup
    # configure_logging()
    pass


def after_all(context):
    logging.info("%s" % ("-" * LINE_LENGTH))
    scenarios = []
    for feature in context._runner.features:
        scenarios += feature.scenarios
    # Scenario.duration is the execution time
    # step.duration is the execution time for each step
    logging.info("PASSED Scenarios:")
    [logging.info("\t" + scenario.name) for scenario in scenarios if scenario.status.name == PASSED]
    logging.info("FAILED Scenarios:")
    [logging.info("\t" + scenario.name) for scenario in scenarios if scenario.status.name == FAILED]
    logging.info("")
    logging.info("SLAYER EXECUTION FINISHED")


def before_tag(context, tag):
    pass


def after_tag(context, tag):
    pass


def before_feature(context, feature):
    logging.info("%s" % ("-" * LINE_LENGTH))
    logging.info("Feature '{}' Start".format(feature.name.upper()))
    logging.info("%s" % ("-" * LINE_LENGTH))


def after_feature(context, feature):
    logging.info("Feature '{}' End".format(feature.name.upper()))
    logging.info("%s" % ("-" * LINE_LENGTH))


def before_scenario(context, scenario):
    logging.info("Scenario '{}' Start".format(scenario.name.upper()))


def after_scenario(context, scenario):
    logging.info("Scenario '{0}' End. Result: '{1}'".format(scenario.name.upper(), scenario.status.name.upper()))
    logging.info("%s" % ("-" * LINE_LENGTH))


def before_step(context, step):
    logging.info("Step: '{}'   {}".format(step.name, "# {}:{}".format(step.location.filename, step.location.line)))


def after_step(context, step):
    if step.status.name == FAILED:
        logging.info("Step: '{}': {}".format(step.name, step.status.name.upper()))



import logging
from abc import abstractmethod

from selenium import webdriver
from selenium.common.exceptions import NoSuchElementException, TimeoutException
from selenium.webdriver.common.by import By
from selenium.webdriver.support.ui import WebDriverWait


class PageObject(object):
    """Base class for Page objects for web_1-based applications"""

    url = None

    # TODO: Handle StaleElementException problems
    # TODO: Find elements. Wait until elem is not present
    def __init__(self, webdriver, url=None, timeout=30):
        self.webdriver = webdriver
        self.url = url
        self.timeout = timeout
        self.webdriver.set_page_load_timeout(timeout)

    def navigate(self):
        self.webdriver.get(self.url)
        return self.validate_page()

    @abstractmethod
    def validate_page(cls):
        return

    def find_element(self, locator, element, wait=3, retries=3):
        """Find element on page.
        Function retries @retries times, waiting @wait seconds after each iteration"""
        result = None
        for _ in range(retries):
            try:
                result = WebDriverWait(self.webdriver, wait).until(lambda x: x.find_element(locator, element))
                logging.debug("Element '{}' found".format(element))
                break
            except (NoSuchElementException, TimeoutException):
                logging.debug("Element could not be found. Retrying...".format(element))
            except Exception as ex:
                logging.exception("An error occurred when trying to find element with ID '{}'".format(element))
                raise
        else:
            logging.debug("Element '{}':'{}' could not be found".format(locator, element))
        return result

    def find_elements(self, by_locator, locator_value, wait=3, retries=3):
        """Function retries @retries times, waiting @wait seconds after each iteration"""
        # for _ in retries:
        raise NotImplemented


if __name__ == "__main__":
    # TODO: Delete
    driver = webdriver.Chrome()
    pp = PageObject(driver)
    pp.webdriver.get("https://www.google.com")
    print(pp.find_element(By.ID, "lst-ib"))
    print(pp.find_element(By.CLASS_NAME, "gsfi"))
    print(pp.find_element(By.XPATH, '//*[@id="lst-ib"]'))
    print(pp.find_element(By.LINK_TEXT, "Google Search"))
    pp.webdriver.close()



from setuptools import setup, find_packages
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='slayer',
    version='0.1.0a11',

    description='Slayer... the QA Automation Framework that came to SLAY!',
    long_description=long_description,
    long_description_content_type='text/x-rst',

    url='https://github.com/FrancoLM/slayer',
    author='Franco Martinez',
    author_email='martinez.franco.leonardo@gmail.com',

    classifiers=[
        # How mature is this project?
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 3 - Alpha',

        'Intended Audience :: Developers',
        'Topic :: Software Development :: Quality Assurance',

        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.6'
    ],
    license="BSD",
    keywords='slayer automation framework',  # Optional
    install_requires=[
        "behave >= 1.2.6",
        "certifi >= 2018.4.16",
        "chardet >= 3.0.4",
        "configobj >= 5.0.6",
        "future >= 0.16.0",
        "idna >= 2.6",
        "parse >= 1.8.2",
        "parse-type >= 0.4.2",
        "PyYAML >= 3.12",
        "requests >= 2.18.4",
        "six >= 1.11.0",
        "urllib3 >= 1.22",
    ],
    packages=find_packages(exclude=['contrib', 'docs', 'tests', 'output', 'deployment', 'tutorial']),  # Required
    include_package_data=True,
)



import logging.config
import os
import shutil
import sys
from argparse import ArgumentParser

import yaml
from behave.formatter.base import StreamOpener
from yaml.scanner import ScannerError

from slayer.slayer_configuration import get_slayer_configuration
from slayer.slayer_logger import SysStdoutLogger


class SlayerRunner(object):
    """Slayer framework object"""

    def __init__(self, framework_name):
        self.variables = {}
        self.framework_name = framework_name
        self.arguments = None

    def log_slayer_information(self):
        self.log_banner()
        print("")
        self.log_environment_variables()
        print("\n")


    def log_banner(self):
        """Prints the Framework Banner in console"""
        dashes = 50
        # FIXME: add your own name
        print("{} FRAMEWORK".format(self.framework_name).center(dashes, "-"))
        print("-" * dashes)

    @classmethod
    def print_variable_name_and_value(cls, name, value):
        """Prints variables in the execution console"""
        print("{var_name:<24} ==> {var_value}".format(var_name=name, var_value=value))

    def configure_stdout_logger(self):
        """Configures the sys.stdout output in order to replicate the console output to a file"""
        sys.stdout = SysStdoutLogger(path=self.variables["FWK_LOGS_DIR"], filename=self.variables["FWK_LOG_FILENAME"])

    def log_environment_variables(self):
        """Prints the Slayer environment variables to console"""
        for name, value in self.variables.items():
            if value:
                self.print_variable_name_and_value(name, value)

    def set_new_environment_variable(self, name, value, print_to_console=False):
        """Creates a new environment variable and stores it in the variables attribute.

        Keyword arguments:
        name -- the name of the new environment variable
        value -- the value it will be assigned
        print_to_console -- whether to print the variable in the console (default False)
        """
        try:
            os.environ[name] = value
            self.variables[name] = value
            if print_to_console:
                self.print_variable_name_and_value(name, value)
            return self.variables[name]
        except TypeError:
            print("ERROR. Environment variable {} could not be set!".format(name))
            raise

    def overwrite_default_configuration(self):
        # TODO
        pass

    def generate_allure_report(self):
        pass

    def parse_slayer_arguments(self):
        """Reads the Slayer-specific arguments provided in the execution of Slayer."""
        parser = ArgumentParser(description='Slayer Framework... it came to SLAY!')
        parser.add_argument('--slayer-config',
                            required=False,
                            action='store',
                            help='Framework Configuration File',
                            default='{}{}config{}config.cfg'.format(os.path.dirname(__file__), os.sep, os.sep))

        parser.add_argument('--logs-config',
                            required=False,
                            help='Framework Loggers settings',
                            default='{}{}config{}logger.yaml'.format(os.path.dirname(__file__), os.sep, os.sep))

        parser.add_argument('--log-filename',
                            required=False,
                            help='Framework Logs Configuration File name',
                            default='output.log')

        parser.add_argument('--behave-config',
                            required=False,
                            help='Relative Path for the Behave Configuration File. The file must be named "behave.ini"',
                            default='')

        default_args, other_args = parser.parse_known_args()
        self.arguments = default_args

        # Remove the custom parameters from sys.argv
        sys.argv = sys.argv[:1] + other_args

    def set_environment_variables(self):
        """Sets the value for all the Slayer-related environment variables"""
        args = self.arguments

        # Set the environment FWK_ROOT to the path where the main script is executed. This is to provide
        # Slayer with the ability of using custom config and feature files and modify behavior mid-execution
        self.set_new_environment_variable("FWK_ROOT", os.getcwd())

        slayer_config = os.path.join(self.variables["FWK_ROOT"], args.slayer_config)
        self.set_new_environment_variable("FWK_CONFIG", slayer_config)

        logs_config = os.path.join(self.variables["FWK_ROOT"], args.logs_config)
        self.set_new_environment_variable("LOGS_CONFIG", logs_config)

        # Get the configuration options from the Slayer config file
        cfg = get_slayer_configuration()
        slayer_cfg = cfg["slayer"]

        slayer_output_dir = os.path.join(self.variables["FWK_ROOT"], slayer_cfg["output"]["path"])
        self.set_new_environment_variable("FWK_OUTPUT_DIR", slayer_output_dir)

        slayer_logs_dir = os.path.join(self.variables["FWK_OUTPUT_DIR"], slayer_cfg["logs"]["path"])
        self.set_new_environment_variable("FWK_LOGS_DIR", slayer_logs_dir)

        slayer_screenshots_dir = os.path.join(self.variables["FWK_OUTPUT_DIR"], slayer_cfg["artifacts"]["path"])
        self.set_new_environment_variable("FWK_ARTIFACTS_DIR", slayer_screenshots_dir)

        slayer_log_filename = os.path.join(slayer_logs_dir, args.log_filename)
        self.set_new_environment_variable("FWK_LOG_FILENAME", slayer_log_filename)

        proxy = slayer_cfg["proxy"]
        self.set_new_environment_variable("HTTP_PROXY", proxy["http_proxy"])
        self.set_new_environment_variable("HTTPS_PROXY", proxy["https_proxy"])
        self.set_new_environment_variable("NO_PROXY", proxy["no_proxy"])

        # APPDATA is used by the Behave executor
        app_data = os.path.join(self.variables["FWK_ROOT"], args.behave_config)
        self.set_new_environment_variable("APPDATA", app_data)

    def cleanup_output_folder(self):
        """Creates and empties the folder where the output artifacts and logs will be stored after executing Slayer"""
        self.delete_output_folders()
        output_folder = self.variables["FWK_OUTPUT_DIR"]
        logs_folder = self.variables["FWK_LOGS_DIR"]
        artifacts_folder = self.variables["FWK_ARTIFACTS_DIR"]
        for folder in (output_folder, logs_folder, artifacts_folder):
            if not os.path.isdir(folder):
                os.makedirs(folder)

    def delete_output_folders(self):
        """Cleans the output folder, where the logs and results of the execution are stored."""
        shutil.rmtree(self.variables["FWK_OUTPUT_DIR"], ignore_errors=True)

    def configure_execution(self):
        self.configure_logging()

    def configure_logging(self):
        """Reads the logger configuration file and set the logger for Slayer.

        Function sets all config-related settings, like log-level and format. If the config file cannot be found,
        then the default logger file is used
        """
        try:
            with open(self.variables["LOGS_CONFIG"], 'r') as f:
                log_config = yaml.safe_load(f.read())
            if "filename" in log_config["handlers"]["file"].keys():
                filename = self.arguments.log_filename or log_config["handlers"]["file"]["filename"]
                log_config["handlers"]["file"]["filename"] = os.path.join(os.getenv("FWK_LOGS_DIR"), filename)
            logging.config.dictConfig(log_config)
        except KeyError:
            print("Could not load logging settings. Using default configuration")
        except ScannerError:
            print("There was an error when loading the logging configuration")
            raise
